/*
 * System Interface Library for games
 * Copyright (c) 2007-2016 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/sysdep/windows/mutex.c: Mutex routines for Windows.
 */

#define IN_SYSDEP

#include "src/base.h"
#include "src/memory.h"
#include "src/sysdep.h"
#include "src/sysdep/windows/internal.h"

/*************************************************************************/
/************************** Interface routines ***************************/
/*************************************************************************/

SysMutexID sys_mutex_create(UNUSED int recursive, int initially_locked)
{
    CRITICAL_SECTION *mutex = mem_alloc(sizeof(*mutex), 0, 0);
    if (UNLIKELY(!mutex)) {
        DLOG("Failed to allocate memory for mutex");
        return NULL;
    }
    InitializeCriticalSection(mutex);
    if (initially_locked) {
        EnterCriticalSection(mutex);
    }
    return (SysMutexID)mutex;
}

/*-----------------------------------------------------------------------*/

void sys_mutex_destroy(SysMutexID mutex)
{
    mem_free((CRITICAL_SECTION *)mutex);
}

/*-----------------------------------------------------------------------*/

int sys_mutex_lock(SysMutexID mutex, float timeout)
{
    if (timeout < 0) {
        EnterCriticalSection((CRITICAL_SECTION *)mutex);
        return 1;
    } else {
        DWORD timeout_ms = timeout_to_ms(timeout);
        DWORD now = GetTickCount();
        while (timeout_ms > 0) {
            if (TryEnterCriticalSection((CRITICAL_SECTION *)mutex)) {
                return 1;
            }
            Sleep(1);
            const DWORD last_iter = now;
            now = GetTickCount();
            const DWORD delta = now - last_iter;
            if (delta >= timeout_ms) {
                timeout_ms = 0;
            } else {
                timeout_ms -= delta;
            }
        }
        return TryEnterCriticalSection((CRITICAL_SECTION *)mutex);
    }
}

/*-----------------------------------------------------------------------*/

void sys_mutex_unlock(SysMutexID mutex)
{
    LeaveCriticalSection((CRITICAL_SECTION *)mutex);
}

/*************************************************************************/
/*************************************************************************/

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
