/*
 * System Interface Library for games
 * Copyright (c) 2007-2016 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/test/time.c: Tests for the high-level timekeeping functions.
 */

#include "src/base.h"
#include "src/sysdep.h"
#include "src/sysdep/test.h"
#include "src/test/base.h"
#include "src/time.h"

/*-----------------------------------------------------------------------*/

/* Utility macro to check time values and generate useful failure messages. */
#define CHECK_THAT_TIME_IS(expected_time)  do {                 \
    const double _expected_time = (expected_time);              \
    const double _time_now = time_now();                        \
    if (_time_now != _expected_time) {                          \
        FAIL("Expected time %g (0x%016llX), got %g"             \
             " (0x%016llX) for time_now()",                     \
             _expected_time, DOUBLE_BITS(_expected_time),       \
             _time_now, DOUBLE_BITS(_time_now));                \
    }                                                           \
} while (0)

/*************************************************************************/
/*************************************************************************/

DEFINE_GENERIC_TEST_RUNNER(test_time)

TEST_INIT(init)
{
    time_init();
    return 1;
}

TEST_CLEANUP(cleanup)
{
    /* Re-init (since there's no time_cleanup()) to make sure we don't
     * leave any junk behind. */
    time_init();
    time_mark_reset();
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_time_now)
{
    sys_test_time_set_seconds(1.0);
    CHECK_THAT_TIME_IS(1.0);
    sys_test_time_set_seconds(1.5);
    CHECK_THAT_TIME_IS(1.5);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_time_delay)
{
    sys_test_time_set_seconds(0.0);
    time_delay(1.0);
    CHECK_THAT_TIME_IS(1.0);
    sys_test_time_set_seconds(0.0);
    time_delay(1.0 + 1.0/(1ULL<<52));
    CHECK_THAT_TIME_IS(1.0 + 1.0/sys_time_unit());
    sys_test_time_set_seconds(0.0);
    time_delay(120.0f);  // Longer than the 60sec limit on sys_time_delay().
    CHECK_THAT_TIME_IS(120.0);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_time_delay_until)
{
    sys_test_time_set_seconds(1.0);
    time_delay_until(1.5);
    CHECK_THAT_TIME_IS(1.5);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_time_mark)
{
    sys_test_time_set_seconds(0.0);
    time_mark_reset();
    time_delay(1.0f);
    time_mark(123);
    time_delay(1.0f);
    time_mark(456);
    time_delay(1.0f);
    CHECK_DOUBLEEQUAL(time_get_mark(123), 1.0);
    CHECK_DOUBLEEQUAL(time_get_mark(456), 2.0);
    CHECK_DOUBLEEQUAL(time_get_mark(789), -1.0);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_time_mark_overflow)
{
    sys_test_time_set_seconds(0.0);
    time_mark_reset();
    for (int i = 1; i <= TIME_MAX_MARKS+1; i++) {
        time_delay(1.0f);
        time_mark(i);
    }
    for (int i = 1; i <= TIME_MAX_MARKS; i++) {
        CHECK_DOUBLEEQUAL(time_get_mark(i), i);
    }
    CHECK_DOUBLEEQUAL(time_get_mark(TIME_MAX_MARKS+1), -1.0);

    return 1;
}

/*************************************************************************/
/*************************************************************************/

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
